/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifndef COMMON_H
#define COMMON_H  1

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <alsa/asoundlib.h>
#include <gdk/gdkkeysyms.h>
#include <math.h>

/* General definitions for signal processing */
#define BAUD_RATE			31.25	/* PSK baude rate */

/* Size of char arrays (strings) for error messages etc */
#define MESG_SIZE	128

/* Frequency in Hz of phase detector (BFO pitch) */
#define AUDIO_FREQUENCY		500

#define SQUELCH_MULTIPLIER	512

/* Control flags */
#define TRANSMIT_TAG		0x00000001 /* Send tag string  */
#define SQUELCH_OPEN		0x00000002 /* Squelch is open  */
#define MODE_BPSK			0x00000004 /* Mode is BPSK     */
#define MODE_QPSK_USB		0x00000008 /* Mode is QPSK-USB */
#define MODE_QPSK_LSB		0x00000010 /* Mode is QPSK-LSB */
#define ENABLE_CAT			0x00000020 /* Enable Trcvr CAT */
#define CAT_SETUP			0x00000040 /* Trcvr CAT setup OK  */
#define ENABLE_RTS_DTR		0x00000080 /* Enable RTS control of PTT */
#define ENABLE_DTR			0x00000100 /* Enable DTR control of PTT */
#define RTS_DTR_SETUP		0x00000200 /* RTS/DTR control setup OK */
#define RECORD_QSO			0x00000400 /* Enable recording of QSO's   */
#define RECEIVE_MODE		0x00000800 /* xpsk31 is in Receive mode   */
#define TRANSMIT_MODE		0x00001000 /* xpsk31 is in Transmit mode  */
#define TRANSMIT_MACRO		0x00002000 /* Start transmission of Macro */
#define TRANSMIT_KEYBD		0x00004000 /* Transmission from Keyboard  */
#define CAPITAL_LETTERS		0x00008000 /* Capitalize letters in Trans */
#define CHANNEL_A_SELECT	0x00010000 /* Detector channel A selected */
#define CAPTURE_SETUP		0x00020000 /* Sound Capture has been set-up  */
#define PLAYBACK_SETUP		0x00040000 /* Sound Playback has been set-up */
#define MIXER_SETUP			0x00080000 /* Sound Mixer has been set-up */
#define SAVE_RECORD			0x00100000 /* Record was saved to file    */
#define ENABLE_AFC			0x00200000 /* Enable Auto Freq. Control   */
#define ENABLE_MAGNIPHASE	0x00400000 /* Enable display of magniphase */
#define TRANSMIT_MORSE_ID	0x00800000 /* Enable display of magniphase */
#define TCVR_SERIAL_TEST	0x01000000 /* Transceiver serial port under test */

/* Some character definitions */
#define HT				0x09 /* Horizontal Tab  */
#define CR				0x0d /* Carriage Return */
#define LF				0x0a /* Line Feed       */
#define BS				0x08 /* Back Space      */
#define NO_CHARACTER	0x80 /* No psk31 character decoded yet */

#define NUM_VARICODE_CHARS  128

/* Return values */
#define SUCCESS		0
#define ERROR		1

/* For the Error Dialog */
#define QUIT	TRUE
#define OK		FALSE

/* 'NUL' character used as 'last character' flag */
#define NULL_CHAR	0x00

/* Number and size of labels for commands  */
#define NUM_OF_LABELS	10
#define LABEL_LENGTH	24

/* Degrees-to-rads and 2*pi */
#define DEG2RAD	0.0174532

#ifndef M_2PI
  #define M_2PI		6.28318530717958647692
#endif

/*-------------------------------------------------------------------*/

/* Configuration parameters structure */
typedef struct
{
  char
	snd_card[32], /* Sound card name */
	pcm_dev[32],  /* PCM device name */
	cap_src[32],  /* Capture source  */
	cap_vol[32],  /* Capture volume control  */
	mst_vol[32],  /* Master volume control   */
	pcm_vol[32];  /* Playback volume control */

  int
	channel,	/* ALSA "channel" for use in mixer setup */
	num_chn,	/* Number of audio channels (2=stereo, 1=mono) */
	use_chn,	/* Channel in use (frontleft=0, frontright=1 etc) */
	cap_lev,	/* Recording/Capture level */
	pbk_lev,	/* Playback (volume) level */
	tx_output,	/* Tx output power level */
	dsp_rate;	/* DSP rate (speed) samples/sec */

  /* Four quadrants of a signal cycle (in dsp samples) */
  double
	quadrant_1,
	quadrant_2,
	quadrant_3,
	quadrant_4;

  /* Transceiver type 1=FT847 2=FT857 3=K2 4=K3 0=NONE */
  int tcvr_type;

  char
	cat_serial[32], /* CAT serial port */
	rts_serial[32]; /* RTS serial port */

  char font[32]; /* Font to use in text views */

  int
	sqlch_thr,		/* Squelch threshold    */
	tone_dur,		/* Preamble tone duration   */
	num_rev,		/* Pre/post-amble reversals */
	wpm,			/* Morse encoder speed: wpm */
	wwrap,			/* Word wrap column number  */
	psk_elem2,		/* Duration of half PSK element in dsp samples */
	psk_elem,		/* Duration of a PSK element in dsp samples */
	tone_period,	/* Period of audio tone in dsp cycles */
	ifft_stride,	/* IFFT stride over input data (dsp samples) */
	rev_magniphase;	/* Reverse the sense of the magniphase indicators */

  /* Local operator data */
  char
	call[15],
	name[13],
	zone[15],
	qth[13],
	loc[7];

  /* Tx/Rx power, N.Fig, antennas */
  char
	tx[12],
	tx_power[10],
	tx_ant[16],
	rx[12],
	rx_nfig[8],
	rx_ant[16];

	/* Xpsk31 glade file path */
	char xpsk31_glade[128];

} rc_data_t;

/* QSO record type definition */
typedef struct
{
  /* QSO data */
  char
	dx_call[15],  /* Remote (dx) station callsign */
	dx_rst[4],    /* My RST report to Dx station  */
	my_rst[4],    /* My RST report to Dx station  */
	dx_name[13],  /* Remote station operator name */
	dx_qth[13],   /* Remote station QTH loc. name */
	dx_loc[7],    /* Remote station QTH locator   */
	date[12],     /* Normal format QSO date  */
	date_adif[9], /* QSO date in ADIF format */
	time[6],      /* Normal format QSO time  */
	time_adif[5], /* QSO time in ADIF format */
	freq[14],     /* QSO operating frequency MHz */
	mode[12],     /* Mode of operation */
	via[13];      /* Via which path or relay Sat */

  /* Remarks on QSO */
  char remarks[46];

} qso_record_t;

/*-------------------------------------------------------------------*/

/* Transceiver status data */
typedef struct
{
  int
	comp_lev,	/* Compressor Level */
	tx_power,	/* Tx Power Output  */
	vfo_freq,	/* VFO A frequency  */
	rig_mode;	/* Transceiver mode (usb|cw etc) */

} tcvr_status_t;

/*-------------------------------------------------------------------*/

/*
 * Standard gettext macros.
 */
#ifdef ENABLE_NLS
#  include <libintl.h>
#  undef _
#  define _(String) dgettext (PACKAGE, String)
#  define Q_(String) g_strip_context ((String), gettext (String))
#  ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#  else
#    define N_(String) (String)
#  endif
#else
#  define textdomain(String) (String)
#  define gettext(String) (String)
#  define dgettext(Domain,Message) (Message)
#  define dcgettext(Domain,Message,Type) (Message)
#  define bindtextdomain(Domain,Directory) (Domain)
#  define _(String) (String)
#  define Q_(String) g_strip_context ((String), (String))
#  define N_(String) (String)
#endif

/*-------------------------------------------------------------------*/

/* Function prototypes produced by cproto */
/* callbacks.c */
void Error_Dialog(char *mesg, gboolean hide);
void Quit_Dialog(char *mesg);
void Save_Dialog(char *mesg);
void Bad_Entry_Dialog(char *mesg);
void on_main_window_destroy(GObject *object, gpointer user_data);
gboolean on_main_window_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);
gboolean on_tx_textview_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
gboolean on_rx_textview_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
void on_squelch_value_changed(GtkRange *range, gpointer user_data);
void on_op_name_changed(GtkEditable *editable, gpointer user_data);
void on_qth_name_changed(GtkEditable *editable, gpointer user_data);
void on_locator_changed(GtkEditable *editable, gpointer user_data);
void on_band_changed(GtkEditable *editable, gpointer user_data);
void on_callsign_changed(GtkEditable *editable, gpointer user_data);
void on_new_record_clicked(GtkButton *button, gpointer user_data);
void on_save_record_clicked(GtkButton *button, gpointer user_data);
void on_clear_fields_clicked(GtkButton *button, gpointer user_data);
void on_error_quit_button_clicked(GtkButton *button, gpointer user_data);
void on_cancel_save_clicked(GtkButton *button, gpointer user_data);
void on_save_ok_clicked(GtkButton *button, gpointer user_data);
void on_bpsk_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_qpsk_lsb_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_qpsk_usb_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_capitalize_letters_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_quit_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_record_qsos_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_identify_in_cw_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_bad_entry_ok_clicked(GtkButton *button, gpointer user_data);
void on_cancel_quit_clicked(GtkButton *button, gpointer user_data);
void on_quit_ok_clicked(GtkButton *button, gpointer user_data);
void on_clear_window_activate(GtkMenuItem *menuitem, gpointer user_data);
gboolean on_tx_textview_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data);
gboolean on_scope_draw(GtkWidget *widget, cairo_t *cr, gpointer user_data);
gboolean on_scope_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
void on_clear_windows_clicked(GtkButton *button, gpointer user_data);
void on_freq_adjust_value_changed(GtkRange *range, gpointer user_data);
void on_k2_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_k3_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_ft847_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_ft857_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_rts_ctl_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_none_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_xmit_clicked(GtkButton *button, gpointer user_data);
void on_rcve_clicked(GtkButton *button, gpointer user_data);
void on_rst_out_changed(GtkEditable *editable, gpointer user_data);
void on_rst_in_changed(GtkEditable *editable, gpointer user_data);
void on_macro1_clicked(GtkButton *button, gpointer user_data);
void on_macro2_clicked(GtkButton *button, gpointer user_data);
void on_macro3_clicked(GtkButton *button, gpointer user_data);
void on_macro4_clicked(GtkButton *button, gpointer user_data);
void on_macro5_clicked(GtkButton *button, gpointer user_data);
void on_macro6_clicked(GtkButton *button, gpointer user_data);
void on_macro7_clicked(GtkButton *button, gpointer user_data);
void on_macro8_clicked(GtkButton *button, gpointer user_data);
void on_macro9_clicked(GtkButton *button, gpointer user_data);
void on_macro10_clicked(GtkButton *button, gpointer user_data);
void on_error_dialog_destroy(GObject *object, gpointer user_data);
void on_error_ok_button_clicked(GtkButton *button, gpointer user_data);
gboolean on_error_dialog_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);
gboolean on_bad_entry_dialog_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);
void on_scope_configure_event(GtkWidget *widget, GdkEventConfigure *event, gpointer user_data);
void on_txvol_spinbutton_value_changed(GtkSpinButton *spinbutton, gpointer user_data);
void on_rxvol_spinbutton_value_changed(GtkSpinButton *spinbutton, gpointer user_data);
void on_afc_checkbutton_toggled(GtkToggleButton *togglebutton, gpointer user_data);
void on_scope_button_clicked(GtkButton *button, gpointer user_data);
/* cat.c */
gboolean Open_Tcvr_Serial(void);
void Close_Tcvr_Serial(void);
gboolean Open_RTS_Serial(void);
gboolean Read_Rx_Status(char *freq, char *rst);
gboolean Set_Tcvr_Mode(void);
gboolean Tune_Tcvr(double x);
gboolean Set_Tx_Power(int tx_power, int comp_lev);
/* display.c */
gboolean Create_Scope_Pixbuf(int width, int height);
void Display_Waterfall(void);
void Set_TxRx_Labels(void);
void Clear_Record_Fields(gboolean all);
void Print_Character(int printchr, GtkTextBuffer *text_buffer, GtkWidget *scrolledwindow, GtkTextIter *iter);
/* ifft.c */
gboolean Initialize_IFFT(int16_t width);
void IFFT_Real(int16_t *data );
void IFFT_Data(int16_t sample);
/* interface.c */
GtkWidget *Builder_Get_Object(GtkBuilder *builder, gchar *name);
GtkWidget *create_main_window(GtkBuilder **builder);
GtkWidget *create_error_dialog(GtkBuilder **builder);
GtkWidget *create_save_dialog(GtkBuilder **builder);
GtkWidget *create_popup_menu(GtkBuilder **builder);
GtkWidget *create_bad_entry_dialog(GtkBuilder **builder);
GtkWidget *create_quit_dialog(GtkBuilder **builder);
/* main.c */
int main(int argc, char *argv[]);
void Usage(void);
/* operation.c */
gboolean Transmit_Keybd(gpointer data);
gboolean Transmit_Macro(gpointer data);
gboolean Receive_Mode(gpointer data);
gboolean Oen_Record_File(void);
/* phase.c */
gboolean PSK31_Phase(int *phase_chg);
void Update_Magniphase(cairo_t *cr);
/* pskcodec.c */
gboolean Decode_PSK_Character(int *dec_char);
gboolean Transmit_Preamble(void);
gboolean Transmit_Postamble(void);
gboolean Transmit_Element(int phase_shift);
gboolean Transmit_PSK_Character(int ascii_chr);
void Init_Viterbi(void);
gboolean Morse_Transmit(char *mesg);
gboolean Make_Cosine_Table(void);
gboolean Make_Sine_Table(void);
/* shared.c */
/* sound.c */
gboolean Open_Capture(char *mesg, int *error);
gboolean Open_Playback(char *mesg, int *error);
gboolean Set_Playback_Level(int level, char *mesg, int *error);
gboolean Set_Capture_Level(int level, char *mesg, int *error);
void Close_Playback(void);
void Close_Capture(void);
void Close_Mixer(void);
gboolean Signal_Sample(short *sample_val);
gboolean DSP_Write(short *buffer, int buff_size);
/* utils.c */
gboolean Load_Config(gpointer data);
gboolean Save_QSO_Record(void);
void Process_RST(GtkEditable *editable);
void Select_Macro(int macro_num);
gboolean Read_QSO_Record(void);
gboolean mem_alloc(void **ptr, size_t req);
gboolean mem_realloc(void **ptr, size_t req);
void free_ptr(void **ptr);
void Strlcpy(char *dest, const char *src, size_t n);
void Strlcat(char *dest, const char *src, size_t n);
void Cleanup(void);
int isFlagSet(int flag);
int isFlagClear(int flag);
void SetFlag(int flag);
void ClearFlag(int flag);
void ToggleFlag(int flag);

#endif

