/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "display.h"
#include "shared.h"

/*------------------------------------------------------------------------*/

  static int
IFFT_Bin_Value( int sum_i, int sum_q, gboolean reset )
{
 /* Value of ifft output "bin" */
  static int bin_val = 0;

  /* Maximum value of ifft bins */
  static int bin_max = 1000, max = 0;


  /* Scale output from IFFT */
  sum_i /= IFFT_SCALE;
  sum_q /= IFFT_SCALE;

  /* Calculate sliding window average of max bin value */
  if( reset )
  {
	bin_max = max;
	if( !bin_max ) bin_max = 1;
	max = 0;
  }
  else
  {
	/* Calculate average signal power at each frequency (bin) */
	bin_val  = bin_val * AMPL_AVE_MUL;
	bin_val += sum_i * sum_i + sum_q * sum_q;
	bin_val /= AMPL_AVE_WIN;

	/* Record max bin value */
	if( max < bin_val ) max = bin_val;

	/* Scale bin values to 255 depending on max value */
	int ret = ( 255 * bin_val ) / bin_max;
	if( ret > 255 ) ret = 255;
	return( ret );
  }

  return( 0 );
} /* IFFT_Bin_Value() */

/*------------------------------------------------------------------------*/

/* Create_Scope_Pixbuf()
 *
 * Creates a pixbuf for the scope display
 */
  gboolean
Create_Scope_Pixbuf( int width, int height )
{
  int idx;

  /* Destroy existing pixbuff */
  if( wfall_pixbuf != NULL )
  {
	g_object_unref( G_OBJECT(wfall_pixbuf) );
	wfall_pixbuf = NULL;
  }

  /* Create waterfall pixbuf */
  wfall_pixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, FALSE, 8, width, height );
  if( wfall_pixbuf == NULL ) return( FALSE );

  wfall_pixels = gdk_pixbuf_get_pixels( wfall_pixbuf );
  wfall_width  = gdk_pixbuf_get_width ( wfall_pixbuf );
  wfall_height = gdk_pixbuf_get_height( wfall_pixbuf );
  wfall_rowstride  = gdk_pixbuf_get_rowstride( wfall_pixbuf );
  wfall_n_channels = gdk_pixbuf_get_n_channels( wfall_pixbuf );
  gdk_pixbuf_fill( wfall_pixbuf, 0 );

  /* Allocate average bin value buffer */
  if( !mem_realloc((void **)&bin_ave,
		(size_t)wfall_width * sizeof(int)) )
	return( FALSE );

  for( idx = 0; idx < wfall_width; idx++ )
	bin_ave[idx] = 0;

  /* Initialize ifft. The waterfall width has to be odd
   * to provide a center line and the FFT width has to
   * be a power of 2 so we compensate by adding 1 */
  if( ! Initialize_IFFT((int16_t)wfall_width + 1) )
	  return( FALSE );

  return( TRUE );

} /* Create_Scope_Pixbuf() */

/*------------------------------------------------------------------------*/

/* Colorize()
 *
 * Pseudo-colorizes FFT Spectrum Display pixels
 * according to the IFFT's output bin values
 */
  static void
Colorize( guchar *pix, int pixel_val )
{
  int n;

  if( pixel_val < 64 ) // From black to blue
  {
	pix[0] = 0;
	pix[1] = 0;
	pix[2] = 3 + (guchar)pixel_val * 4;
  }
  else if( pixel_val < 128 ) // From blue to green
  {
	n = pixel_val - 127; // -63 <= n <= 0
	pix[0] = 0;
	pix[1] = 255 + (guchar)n * 4;
	pix[2] = 3   - (guchar)n * 4;
  }
  else if( pixel_val < 192 ) // From green to yellow
  {
	n = pixel_val - 191; // -63 <= n <= 0
	pix[0] = 255 + (guchar)n * 4;
	pix[1] = 255;
	pix[2] = 0;
  }
  else // From yellow to reddish-orange
  {
	n = pixel_val - 255; // -63 <= n <= 0
	pix[0] = 255;
	pix[1] = 66 - (guchar)n * 3;
	pix[2] = 0;
  }

} /* Colorize() */

/*----------------------------------------------------------------------*/

/* Display_Waterfall()
 *
 * Displays audio spectrum as "waterfall"
 */

  void
Display_Waterfall( void )
{
  int
	idh, idv,	/* Index to hor. and vert. position in warterfall */
	pixel_val,	/* Greyscale value of pixel derived from ifft o/p  */
	idb,		/* Index to average bin values array */
	idf,		/* Index to ifft output array */
	temp;

  /* Constant needed to draw white line in waterfall */
  int white_line = wfall_width / 2;

  /* Pointer to current pixel */
  static guchar *pix;

  /* Draw a vertical white line in waterfall at detector's freq. */
  pix = wfall_pixels + wfall_rowstride + wfall_n_channels;
  pix += wfall_n_channels * white_line;
  pix[0] = pix[1] = pix[2] = 0xff;

  /* Copy each line of waterfall to next one */
  temp = wfall_height - 2;
  for( idv = temp; idv > 0; idv-- )
  {
	pix = wfall_pixels + wfall_rowstride * idv + wfall_n_channels;

	for( idh = 0; idh < wfall_width; idh++ )
	{
	  *pix = *( pix - wfall_rowstride ); pix++;
	  *pix = *( pix - wfall_rowstride ); pix++;
	  *pix = *( pix - wfall_rowstride );
	  pix += wfall_n_channels - 2;
	}
  }

  /* Got to top left of pixbuf */
  pix = wfall_pixels;

  /* First (DC) ifft bin not used */
  bin_ave[0] = 0;

  /* Do the IFFT on input array */
  IFFT_Real( ifft_data );
  idb = 0;
  for( idf = 2; idf < ifft_data_length; idf += 2 )
  {
	/* Calculate power of signal at each freq. ("bin") */
	pixel_val = IFFT_Bin_Value(
		ifft_data[idf], ifft_data[idf + 1], FALSE );

	/* Calculate average bin values */
	bin_ave[idb++] = pixel_val;

	/* Color code signal strength */
	Colorize( pix, pixel_val );
	pix += wfall_n_channels;

  } /* for( idf = 2; idf < ifft_data_length; idf += 2 ) */

  /* Reset function */
  IFFT_Bin_Value( 0, 0, TRUE );

  /* At last draw waterfall */
  gtk_widget_queue_draw( scope );

} /* Display_Waterfall() */

/*------------------------------------------------------------------------*/

/*  Set_TxRx_Labels()
 *
 *  Sets up the labels in the Tx/Rx frames
 */

  void
Set_TxRx_Labels( void )
{
  if( isFlagSet(TRANSMIT_KEYBD) ||
	  isFlagSet(TRANSMIT_MACRO) )
  {
	if( isFlagSet(MODE_BPSK) )
	  gtk_label_set_markup( GTK_LABEL(xmit_status), XMIT_BPSK );
	else if( isFlagSet(MODE_QPSK_USB) )
	  gtk_label_set_markup( GTK_LABEL(xmit_status), XMIT_QPSK_USB );
	else if( isFlagSet(MODE_QPSK_LSB) )
	  gtk_label_set_markup( GTK_LABEL(xmit_status), XMIT_QPSK_LSB );
  }
  else gtk_label_set_markup( GTK_LABEL(xmit_status), XMIT_OFF );

  if( isFlagSet(RECEIVE_MODE) )
  {
	if( isFlagSet(MODE_BPSK) )
	  gtk_label_set_markup( GTK_LABEL(rcve_status), RECV_BPSK );
	else if( isFlagSet(MODE_QPSK_USB) )
	  gtk_label_set_markup( GTK_LABEL(rcve_status), RECV_QPSK_USB );
	else if( isFlagSet(MODE_QPSK_LSB) )
	  gtk_label_set_markup( GTK_LABEL(rcve_status), RECV_QPSK_LSB );
  }
  else gtk_label_set_markup( GTK_LABEL(rcve_status), RECV_OFF );

  /* Close cat if tx/rx off */
  if( isFlagClear(RECEIVE_MODE) &&
	  isFlagClear(TRANSMIT_MACRO) &&
	  isFlagClear(TRANSMIT_KEYBD) )
	Close_Tcvr_Serial();

} /* Set_TxRx_Labels() */

/*------------------------------------------------------------------------*/

/*  Clear_Record_Fields()
 *
 *  Clears all QSO Record fields
 */

  void
Clear_Record_Fields( gboolean all )
{
  gtk_entry_set_text( GTK_ENTRY(
		Builder_Get_Object(main_window_builder, "callsign")), "" );
  gtk_entry_set_text( GTK_ENTRY(
		Builder_Get_Object(main_window_builder, "rst_in")), "" );
  gtk_entry_set_text( GTK_ENTRY(
		Builder_Get_Object(main_window_builder, "rst_out")), "" );
  gtk_entry_set_text( GTK_ENTRY(
		Builder_Get_Object(main_window_builder, "op_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(
		Builder_Get_Object(main_window_builder, "qth_name")), "" );
  gtk_entry_set_text( GTK_ENTRY(
		Builder_Get_Object(main_window_builder, "locator")), "" );
  if( all )
	gtk_entry_set_text( GTK_ENTRY(
		  Builder_Get_Object(main_window_builder, "band")), "" );

  qso_record.dx_call[0] = '\0';
  qso_record.my_rst[0]  = '\0';
  qso_record.dx_rst[0]  = '\0';
  qso_record.dx_qth[0]  = '\0';
  qso_record.dx_loc[0]  = '\0';
  if( all )
	qso_record.freq[0]  = '\0';
  Strlcpy( qso_record.dx_name, "OP", sizeof(qso_record.dx_name) );

  ClearFlag( SAVE_RECORD );
} /*  Clear_Record_Fields() */

/*------------------------------------------------------------------------*/

/*  Print_Character()
 *
 *  Prints a character to a text view port
 */

void
Print_Character(
	int printchr,
	GtkTextBuffer *text_buffer,
	GtkWidget     *scrolledwindow,
	GtkTextIter   *iter )
{
  GtkAdjustment *adjustment;
  char char2text[2];
  static gdouble last_upper = 0.0;

  /* Delete a char on backspace */
  if( printchr == GDK_KEY_BackSpace )
  {
	GtkTextIter start;

	gtk_text_buffer_get_iter_at_offset( text_buffer, &start,
		gtk_text_buffer_get_char_count(text_buffer) -1 );
	gtk_text_buffer_delete( text_buffer, &start, iter );
	return;
  }

  /* Convert char to "text" */
  char2text[0] = (char) printchr;
  char2text[1] = '\0';

  /* Print character */
  gtk_text_buffer_insert( text_buffer, iter, char2text, -1 );

  /* Record to QSO record file */
  if( isFlagSet(RECORD_QSO)    &&
	  (isFlagSet(RECEIVE_MODE) ||
	   isFlagSet(TRANSMIT_MODE)) )
	fprintf( qso_record_fp, "%c", printchr );

  /* Scroll Text View to bottom on LF */
  adjustment = gtk_scrolled_window_get_vadjustment
	( GTK_SCROLLED_WINDOW(scrolledwindow) );
  gdouble upper = gtk_adjustment_get_upper( adjustment );
  gdouble page_size = gtk_adjustment_get_page_size( adjustment );
  if( (upper < last_upper) || (upper > last_upper) )
  {
	gtk_adjustment_set_value( adjustment, upper - page_size );
	last_upper = page_size;
  }

  /* Mark LF in record files */
  if( (printchr == LF) || (printchr == GDK_KEY_Return) )
  {
	if( isFlagSet(RECEIVE_MODE) )
	  fprintf( qso_record_fp, "Rx-> " );
	else if( isFlagSet(TRANSMIT_MODE) )
	  fprintf( qso_record_fp, "Tx-> " );
  }

} /* Print_Character() */

/*------------------------------------------------------------------------*/

