/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "phase.h"
#include "shared.h"

static int
  sel_level_i,   /* In-phase level from selected channel   */
  sel_level_q;   /* Quadrature level from selected channel */

/*------------------------------------------------------------------------*/

/*  Arc_Tan()
 *
 *  Calculates an aproximate integer arc tan (in deg)
 *  for a given pair of integer parameters (y,x). The
 *  result is in range 0 - +/-180 deg, max error ~.5 deg.
 *  The first approximation used is arctan = 45*y/x for
 *  y<x and for y>x arctan = 90-45*x/y. This is used in
 *  a table to improve the accuracy. The signs of x
 *  and y are used to find arc tan in all quadrants.
 */

  static int
Arc_Tan( int y, int x )
{
  int
	angle,   /* Inermediate result  */
	arc_tan; /* Final arctan result */

  /* Arc tan correction table */
  char table[46] = ARC_TAN_TABLE;

  /* Avoid divide by zero */
  if( (x == 0) || (y == 0) ) return( 0 );

  /* Angle +-(0-45) deg or (135-225) */
  if( abs(y) <= abs(x) )
  {
	/* First approximation */
	angle = (45 * y) / x;

	/* Use correction table for more accuracy */
	if( angle < 0 ) angle = -(int)table[-angle];
	else            angle = (int)table[angle];

	/* Set result in correct quadrant */
	if( x < 0 )      arc_tan = 180 + angle;
	else if( y < 0 ) arc_tan = 360 + angle;
	else             arc_tan = angle;

  }  /* if( abs(y) <= abs(x) ) */
  else /* Angle (45-135) or (225-315) */
  {
	/* First approximation */
	angle = (45 * x) / y;

	/* Use correction table for more accuracy */
	if( angle < 0 ) angle = -(int) table[-angle];
	else            angle = (int)table[angle];

	/* Set result in correct quadrant */
	if( y < 0 ) arc_tan = 270 - angle;
	else        arc_tan =  90 - angle;

  } /* else */

  return( arc_tan );
} /* End of Arc_Tan() */

/*------------------------------------------------------------------------*/

/*  Scalar_Magn()
 *
 *  sqrt(x^2 + y^2). This is approximated by a Taylor
 *  series expansion in (x+y) e.g: ((x+y)^2-x*y)/(x+y)
 *  to first order, and refined by one iteration of the
 *  Newton-Raphson method. This is rather wierd but
 *  seems to give the most accurate approximation.
 */

  static int
Scalar_Magn( int x, int y )
{
  /* Vector magnitude of two values at 90 deg */
  int vector_mag;

  int
	sum,    /* Sum of two consecutive samples    */
	prod,   /* Product of two cnsecutive samples */
	aprox;  /* First approxim. to sqrt(x^2+y^2)  */

  /* Remove sign */
  x = abs( x );
  y = abs( y );

  /* Calculate first approximation to sqrt(x^2+y^2) */
  sum   = x + y;
  prod  = x * y;
  aprox = ( sum == 0 ? 0 : (sum * sum - prod) / sum );

  /* Refine by one iteration using Newton-Raphson method */
  vector_mag = ( aprox == 0 ? 0 :
	  ((aprox * aprox + sum * sum - (2 * prod)) / aprox) / 2 );

  return( vector_mag );
} /* End of Scalar_Magn() */

/*------------------------------------------------------------------------*/

/*  PSK31_Phase()
 *
 *  Detects phase changes of the PSK carrier
 */

  gboolean
PSK31_Phase( int *phase_chg )
{
  static double det_phase; /* Phase in samples of phase detector */
  static int
	old_phase,   /* Old carrier phase from detector o/p */
	phase_slip,  /* Average and normalized carrier phase slip */
	chA_squelch, /* Channel A squelch (signal) level */
	chB_squelch, /* Channel B squelch (signal) level */
	last_lev_i,  /* Last phase detector in-phase signal level */
	last_lev_q;  /* Last phase detector quadrature signal level */

  int
	elem_timer,    /* Counter used for producing the baud rate  */
	det_level_i,   /* Phase detector in-phase signal level      */
	det_level_q,   /* Phase detector quadrature signal level    */
	chA_level_i,   /* 'A' channel in-phase signal level         */
	chA_level_q,   /* 'A' channel quadrature signal level       */
	chB_level_i,   /* 'B' channel in-phase signal level         */
	chB_level_q,   /* 'B' channel quadrature signal level       */
	carrier_phase, /* Phase angle of audio signal from receiver */
	max_slip,      /* Maximum phase shift taken as phase slip   */
	slice_cnt;     /* Counts slices (= 1/2 PSK31 elements) done */

	/* Normalized values of detector */
	/* channels and selected squelch */
  int
	chA_norm_lev_i,
	chA_norm_lev_q,
	chB_norm_lev_i,
	chB_norm_lev_q,
	norm_squelch;

  /* Sample level calculated from raw sample */
  short sample_lev;


  /* Initialize levels and psk phase */
  chA_level_i = chA_level_q = 0;
  chB_level_i = chB_level_q = 0;

  /*** Following code acts like a synchronous detector,   ***/
  /*** working at a nominal 500 Hz. To avoid the nulling  ***/
  /*** of the o/p levels when a phase reversal occurs in  ***/
  /*** the middle of a PSK31 element, two 'channels' are  ***/
  /*** used separated by a 1/2 element time delay and the ***/
  /*** one with the best squelch level is selected in use ***/

  /* Return psk phase after two PSK31 'slices' eg 1 element period */
  for( slice_cnt = 0; slice_cnt < 2; slice_cnt++ )
  {
	/* Initialize levels and psk phase */
	det_level_i = det_level_q = 0;

	/* Summate samples for each half element */
	for(elem_timer = 0; elem_timer < rc_data.psk_elem2; elem_timer++ )
	{
	  /* Get signal sample from buffer, abort on error */
	  if( !Signal_Sample(&sample_lev) )
	  {
		ClearFlag( RECEIVE_MODE );
		Close_Capture();
		return( FALSE );
	  }

	  /*** Calculate phase detector in-phase level. ***/
	  /* Samples are accumulated over a PSK31 element */
	  /* From 0 to 180 deg use sample level as is.
	   * From 180 to 360 deg use rectified sample */
	  if( det_phase < rc_data.quadrant_2 )
		det_level_i += sample_lev;
	  else
		det_level_i -= sample_lev;

	  /*** Calculate phase detector quadrature level. ***/
	  /* Samples are accumulated over a PSK31 element   */
	  /* From 90 to 270 deg use sample level as is.
	   * From 270 to 90 deg use rectified sample */
	  if( (det_phase >= rc_data.quadrant_1) &&
		  (det_phase <  rc_data.quadrant_3) )
		det_level_q += sample_lev;
	  else
		det_level_q -= sample_lev;

	  /* Advance detector phase by one DSP sample count */
	  det_phase += 1.0 + freq_error / (double)AUDIO_FREQUENCY;

	  /* At a full signal cycle, reset the */
	  /* phase count of the phase detector */
	  if( det_phase >= rc_data.quadrant_4 )
		det_phase -= rc_data.quadrant_4;

	} /* for(elem_timer = 0; elem_timer < rc_data.psk_elem2;... */

	/* Channel A sums current detector levels */
	chA_level_i += det_level_i;
	chA_level_q += det_level_q;

	/* Channel B sums last detector levels */
	chB_level_i += last_lev_i;
	chB_level_q += last_lev_q;

	/* Save current levels */
	last_lev_i = det_level_i;
	last_lev_q = det_level_q;

  } /* for( slice_cnt = 0; slice_cnt < 2; slice_cnt++ ) */

  /* Normalize detector channel levels */
  chA_norm_lev_i = chA_level_i / rc_data.psk_elem;
  chA_norm_lev_q = chA_level_q / rc_data.psk_elem;
  chB_norm_lev_i = chB_level_i / rc_data.psk_elem;
  chB_norm_lev_q = chB_level_q / rc_data.psk_elem;

  /* Calculate sliding window average values for the channel */
  /* squelch. This is the vector magnitude of in-phase and   */
  /* quadrature signals from the phase detector channels     */
  chA_squelch *= ( SQUELCH_WINDOW - 1 );
  chA_squelch += Scalar_Magn( chA_norm_lev_i, chA_norm_lev_q );
  chA_squelch /= SQUELCH_WINDOW;

  chB_squelch *= ( SQUELCH_WINDOW - 1 );
  chB_squelch += Scalar_Magn( chB_norm_lev_i, chB_norm_lev_q );
  chB_squelch /= SQUELCH_WINDOW;

  /* Set best channel select flag according to squelch */
  if( 4 * chA_squelch > 5 * chB_squelch )
	SetFlag( CHANNEL_A_SELECT );
  else if( 4 * chB_squelch > 5 * chA_squelch )
	ClearFlag( CHANNEL_A_SELECT );

  /* Select best channel data according to flag */
  if( isFlagSet(CHANNEL_A_SELECT) )
  {
	/* Normalize squelch value and select level */
	norm_squelch = chA_squelch;
	sel_level_i  = chA_norm_lev_i;
	sel_level_q  = chA_norm_lev_q;
  }
  else
  {
	/* Normalize squelch value and select level */
	norm_squelch = chB_squelch;
	sel_level_i  = chB_norm_lev_i;
	sel_level_q  = chB_norm_lev_q;
  }

  /* Set squelch open flag */
  if( norm_squelch > rc_data.sqlch_thr )
	SetFlag( SQUELCH_OPEN );
  else
	ClearFlag( SQUELCH_OPEN );

  /* Calculate new carrier phase if squelch is  */
  /* open. Calculate carrier phase from arctan  */
  /* of in-phase and quadrature detector levels */
  if( isFlagSet(SQUELCH_OPEN) )
	carrier_phase = Arc_Tan( sel_level_q, sel_level_i );
  else
	carrier_phase = 0;

  /* Phase change between consecutive phase values */
  *phase_chg = old_phase - carrier_phase;
  old_phase  = carrier_phase;

  /* Correct for discontinuity at +-180 deg */
  if( *phase_chg > 180 )       *phase_chg -= 360;
  else if( *phase_chg < -180 ) *phase_chg += 360;

  /* For BPSK phase slip limit = 90 */
  /* For QPSK phase slip limit = 45 */
  if( isFlagSet(MODE_BPSK) ) max_slip = 90;
  else max_slip = 45;

  /* Limit phase changes to +-max_slip for slip rate calculations. */
  /* Caclulate average phase change with a sliding window averager */
  if( abs(*phase_chg) < max_slip )
  {
	phase_slip *= ( PHASE_AVE_WINDOW - 1 );
	phase_slip += *phase_chg;
	phase_slip /= PHASE_AVE_WINDOW;
  }

  /* Do Auto Freq Control */
  if( isFlagSet(ENABLE_AFC) )
  {
	if( (phase_slip > 1) && (freq_error < MAX_FREQ_ERROR) )
	  freq_error += 0.1;
	else if( (phase_slip < -1) && (freq_error > -MAX_FREQ_ERROR) )
	  freq_error -= 0.1;
  }

  /* Rescale for magniphase display */
  sel_level_i /= MAGNIPHASE_SCALE;
  sel_level_q /= MAGNIPHASE_SCALE;

  /* Plot carrier phase on magniphase */
  if( isFlagSet(ENABLE_MAGNIPHASE) )
	gtk_widget_queue_draw( scope );

  return( TRUE );
} /* End of PSK31_Phase() */

/*------------------------------------------------------------------------*/

/*  Update_Magniphase()
 *
 *  Displays carrier phase and magnitude
 */

  void
Update_Magniphase( cairo_t *cr )
{
  double x, y, e;

  /* Factors needed to draw freq error pointer */
  double
	scope_radius   = (double)wfall_width / 2.0,
	scope_center_x = (double)wfall_width / 2.0,
	scope_center_y = (double)wfall_height / 2.0;

  int
	level_i = sel_level_i,
	level_q = sel_level_q;

  /* Draw a dark green filled circle, background for scope */
  cairo_set_source_rgb( cr, 0.0, .3, 0.0 );
  cairo_arc( cr,
	  scope_center_x,
	  scope_center_y,
	  scope_radius,
	  0.0, M_2PI );
  cairo_fill( cr );

  /* Reverse magniphase sense */
  if( rc_data.rev_magniphase )
	level_q = -level_q;

  /* Draw the signal magniphase vector */
  cairo_set_source_rgb( cr, 0.0, 1.0, 0.0 );
  cairo_move_to( cr, scope_center_x, scope_center_y );
  cairo_line_to( cr,
	  scope_center_x + (double)level_q,
	  scope_center_y + (double)level_i );

  /* Frequency error pointer co-ordinates */
  e = DEG2RAD * freq_error * 18.0; /* Error in rads */
  x = scope_center_x + scope_radius * sin( e );
  y = scope_center_y - scope_radius * cos( e );

  /* Reverse magniphase sense */
  if( rc_data.rev_magniphase ) x = -x;

  /* Draw frequency error pointer line */
  cairo_move_to( cr, scope_center_x, scope_center_y );
  cairo_line_to( cr, x, y );

  /* Stroke paths */
  cairo_stroke( cr );

} /* End of Update_Magniphase() */

/*------------------------------------------------------------------------*/

