/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifndef PSKCODEC_H
#define PSKCODEC_H		1

#include <math.h>
#include "common.h"

/* Trellis data structure for the Viterbi decoder */
typedef struct
{
  int      distance; /* Total excess distance to this state  */
  long int estimate; /* Estimate of transmitted bit sequence */
} trellis_state_t;

/*-------------------------------------------------------------------*/

/*
   The VARICODE Alphabet

   The codes are transmitted left bit first, with '0' representing
   a phase reversal on BPSK and '1' representing a steady carrier.
   A minimum of two zeros is inserted between characters.
   Some implementations may not handle all the codes below 32.
 */

#define VARICODE_ALPHABET \
{ \
  /* ASCII   VARICODE       HEX */ \
  /* NUL   10 1010 1011 */  0x2ab, \
  /* SOH   10 1101 1011 */  0x2db, \
  /* STX   10 1110 1101 */  0x2ed, \
  /* ETX   11 0111 0111 */  0x377, \
  /* EOT   10 1110 1011 */  0x2eb, \
  /* ENQ   11 0101 1111 */  0x35f, \
  /* ACK   10 1110 1111 */  0x2ef, \
  /* BEL   10 1111 1101 */  0x2fd, \
  /* BS    10 1111 1111 */  0x2ff, \
  /* TAB      1110 1111 */  0x0ef, \
  /* LF          1 1101 */  0x01d, \
  /* VT    11 0110 1111 */  0x36f, \
  /* FF    10 1101 1101 */  0x2dd, \
  /* CR          1 1111 */  0x01f, \
  /* SO    11 0111 0101 */  0x375, \
  /* SI    11 1010 1011 */  0x3ab, \
  /* DLE   10 1111 0111 */  0x2f7, \
  /* DC1   10 1111 0101 */  0x2f5, \
  /* DC2   11 1010 1101 */  0x3ad, \
  /* DC3   11 1010 1111 */  0x3af, \
  /* DC4   11 0101 1011 */  0x35b, \
  /* NAK   11 0110 1011 */  0x36b, \
  /* SYN   11 0110 1101 */  0x36d, \
  /* ETB   11 0101 0111 */  0x357, \
  /* CAN   11 0111 1011 */  0x37b, \
  /* EM    11 0111 1101 */  0x37d, \
  /* SUB   11 1011 0111 */  0x3b7, \
  /* ESC   11 0101 0101 */  0x355, \
  /* FS    11 0101 1101 */  0x35d, \
  /* GS    11 1011 1011 */  0x3bb, \
  /* RS    10 1111 1011 */  0x2fb, \
  /* US    11 0111 1111 */  0x37f, \
  /*                  1 */  0x001, \
  /* !      1 1111 1111 */  0x1ff, \
  /* "      1 0101 1111 */  0x15f, \
  /* #      1 1111 0101 */  0x1f5, \
  /* $      1 1101 1011 */  0x1db, \
  /* %     10 1101 0101 */  0x2d5, \
  /* &     10 1011 1011 */  0x2bb, \
  /* '      1 0111 1111 */  0x17f, \
  /* (        1111 1011 */  0x0fb, \
  /* )        1111 0111 */  0x0f7, \
  /* *      1 0110 1111 */  0x16f, \
  /* +      1 1101 1111 */  0x1df, \
  /* ,         111 0101 */  0x075, \
  /* -          11 0101 */  0x035, \
  /* .         101 0111 */  0x057, \
  /* /      1 1010 1111 */  0x1af, \
  /* 0        1011 0111 */  0x0b7, \
  /* 1        1011 1101 */  0x0bd, \
  /* 2        1110 1101 */  0x0ed, \
  /* 3        1111 1111 */  0x0ff, \
  /* 4      1 0111 0111 */  0x177, \
  /* 5      1 0101 1011 */  0x15b, \
  /* 6      1 0110 1011 */  0x16b, \
  /* 7      1 1010 1101 */  0x1ad, \
  /* 8      1 1010 1011 */  0x1ab, \
  /* 9      1 1011 0111 */  0x1b7, \
  /* :        1111 0101 */  0x0f5, \
  /* ;      1 1011 1101 */  0x1bd, \
  /* <      1 1110 1101 */  0x1ed, \
  /* =         101 0101 */  0x055, \
  /* >      1 1101 0111 */  0x1d7, \
  /* ?     10 1010 1111 */  0x2af, \
  /* @     10 1011 1101 */  0x2bd, \
  /* A         111 1101 */  0x07d, \
  /* B        1110 1011 */  0x0eb, \
  /* C        1010 1101 */  0x0ad, \
  /* D        1011 0101 */  0x0b5, \
  /* E         111 0111 */  0x077, \
  /* F        1101 1011 */  0x0db, \
  /* G        1111 1101 */  0x0fd, \
  /* H      1 0101 0101 */  0x155, \
  /* I         111 1111 */  0x07f, \
  /* J      1 1111 1101 */  0x1fd, \
  /* K      1 0111 1101 */  0x17d, \
  /* L        1101 0111 */  0x0d7, \
  /* M        1011 1011 */  0x0bb, \
  /* N        1101 1101 */  0x0dd, \
  /* O        1010 1011 */  0x0ab, \
  /* P        1101 0101 */  0x0d5, \
  /* Q      1 1101 1101 */  0x1dd, \
  /* R        1010 1111 */  0x0af, \
  /* S         110 1111 */  0x06f, \
  /* T         110 1101 */  0x06d, \
  /* U      1 0101 0111 */  0x157, \
  /* V      1 1011 0101 */  0x1b5, \
  /* W      1 0101 1101 */  0x15d, \
  /* X      1 0111 0101 */  0x175, \
  /* Y      1 0111 1011 */  0x17b, \
  /* Z     10 1010 1101 */  0x2ad, \
  /* [      1 1111 0111 */  0x1f7, \
  /* \      1 1110 1111 */  0x1ef, \
  /* ]      1 1111 1011 */  0x1fb, \
  /* ^     10 1011 1111 */  0x2bf, \
  /* _      1 0110 1101 */  0x16d, \
  /* `     10 1101 1111 */  0x2df, \
  /* a             1011 */  0x00b, \
  /* b         101 1111 */  0x05f, \
  /* c          10 1111 */  0x02f, \
  /* d          10 1101 */  0x02d, \
  /* e               11 */  0x003, \
  /* f          11 1101 */  0x03d, \
  /* g         101 1011 */  0x05b, \
  /* h          10 1011 */  0x02b, \
  /* i             1101 */  0x00d, \
  /* j      1 1110 1011 */  0x1eb, \
  /* k        1011 1111 */  0x0bf, \
  /* l           1 1011 */  0x01b, \
  /* m          11 1011 */  0x03b, \
  /* n             1111 */  0x00f, \
  /* o              111 */  0x007, \
  /* p          11 1111 */  0x03f, \
  /* q      1 1011 1111 */  0x1bf, \
  /* r           1 0101 */  0x015, \
  /* s           1 0111 */  0x017, \
  /* t              101 */  0x005, \
  /* u          11 0111 */  0x037, \
  /* v         111 1011 */  0x07b, \
  /* w         110 1011 */  0x06b, \
  /* x        1101 1111 */  0x0df, \
  /* y         101 1101 */  0x05d, \
  /* z      1 1101 0101 */  0x1d5, \
  /* {     10 1011 0111 */  0x2b7, \
  /* |      1 1011 1011 */  0x1bb, \
  /* }     10 1011 0101 */  0x2b5, \
  /* ~     10 1101 0111 */  0x2d7, \
  /* DEL   11 1011 0101 */  0x3b5  \
}

/*-------------------------------------------------------------------*/

/* Polynomials for the convolution encoder */
#define POLYNOM1	0x19
#define POLYNOM2	0x17

/*
   The CONVOLUTIONAL Code

   The left column contains the 32 combinations of a run of five
   Varicode bits, transmitted left bit first. The right column is
   the corresponding phase shift to be applied to the carrier, with
   0 meaning no shift, 1 meaning advance by 90, 2 meaning polarity
   reversal and 3 meaning retard by 90.  A continuous phase advance
   is the same as an HF frequency shift.
 */

#define CONVOLUTIONAL_CODE \
{ \
  /* Bits      Phase */ \
  /* 00000 */  2,       \
  /* 00001 */  1,       \
  /* 00010 */  3,       \
  /* 00011 */  0,       \
  /* 00100 */  3,       \
  /* 00101 */  0,       \
  /* 00110 */  2,       \
  /* 00111 */  1,       \
  /* 01000 */  0,       \
  /* 01001 */  3,       \
  /* 01010 */  1,       \
  /* 01011 */  2,       \
  /* 01100 */  1,       \
  /* 01101 */  2,       \
  /* 01110 */  0,       \
  /* 01111 */  3,       \
  /* 10000 */  1,       \
  /* 10001 */  2,       \
  /* 10010 */  0,       \
  /* 10011 */  3,       \
  /* 10100 */  0,       \
  /* 10101 */  3,       \
  /* 10110 */  1,       \
  /* 10111 */  2,       \
  /* 11000 */  3,       \
  /* 11001 */  0,       \
  /* 11010 */  2,       \
  /* 11011 */  1,       \
  /* 11100 */  2,       \
  /* 11101 */  1,       \
  /* 11110 */  3,       \
  /* 11111 */  0        \
}

/*
   For example, the "space" symbol - a single 1 preceded and followed
   by zeros - would be represented by successive run-of-five groups
   00000, 00001, 00010, 00100, 01000, 10000, 00000.
   The transmitter will send the QPSK pattern .. 2,1,3,3,0,1,2, ..
   Note that a continuous sequence of zeros (the idle sequence)
   gives continuous reversals, the same as BPSK.
 */

/*-------------------------------------------------------------------*/

/* Definitions for ascii to Morse code conversion */
#define NUM_MORSE_CHARS           55 /* Number of chars in Morse table */

/* ASCII equivalents to Morse hex code. Last */
/* one (*) used for unrecognized characters. */
#define ASCII_CHAR \
{ \
  'A','B','C','D','E','F','G','H','I','J','K','L','M',  \
  'N','O','P','Q','R','S','T','U','V','W','X','Y','Z',  \
  '1','2','3','4','5','6','7','8','9','0','.',',',':',  \
  '?','\'','-','/','(','"',';','$','#','<','!','@',']', \
  '=','~',' '										    \
}

/* Hex equivalents to Morse code chars above. Formed   */
/* by starting with a 1 and following with a 0 for     */
/* dash and 1 for dit e.g: A = ditdah = 101 = Hex 0x06 */
#define MORSE_CODE \
{ \
  0x06,0x17,0x15,0x0b,0x03,0x1d,0x09,0x1f,0x07,0x18,0x0a,0x1b,0x04, \
  0x05,0x08,0x19,0x12,0x0d,0x0f,0x02,0x0e,0x1e,0x0c,0x16,0x14,0x13, \
  0x30,0x38,0x3c,0x3e,0x3f,0x2f,0x27,0x23,0x21,0x20,0x6a,0x4c,0x47, \
  0x73,0x61,0x5e,0x2d,0x52,0x6d,0x55,0xf6,0x35,0x7a,0x2a,0x37,0x29, \
  0x2e,0xff,0x01 												    \
}

#define PHASE_SHIFT_0		0		/* PSK phase reference    */
#define PHASE_SHIFT_90		1		/* PSK 90 phase advance   */
#define PHASE_SHIFT_180		2		/* PSK 180 phase reversal */
#define PHASE_SHIFT_270		3		/* PSK 90 phase retard    */

#endif

